// Package websocket...
//
// Description : websocket...
//
// Author : go_developer@163.com<张德满>
//
// Date : 2021-03-27 6:49 下午
package websocket

import (
	"fmt"

	"github.com/go-developer/gopkg/logger"

	"go.uber.org/zap"

	"github.com/go-developer/websocket/message"

	"github.com/go-developer/websocket/config"

	"github.com/tidwall/gjson"

	"github.com/go-developer/websocket/context"

	"gopkg.in/olahol/melody.v1"

	"github.com/gin-gonic/gin"
	"github.com/go-developer/websocket/abstract"
	"github.com/pkg/errors"
)

// Server ...
//
// Author : go_developer@163.com<张德满>
//
// Date : 8:04 下午 2021/3/27
type Server struct {
	ginRouter      *gin.Engine            // GIN引擎
	wsServer       *melody.Melody         // websocket引擎
	conf           *config.WSServerConfig // 配置
	loggerInstance *zap.Logger            // 日志实例
}

var (
	// ginRouterTable 表
	ginRouterTable map[int]*gin.Engine
	// wsServerTable 表
	wsServerTable map[int]map[string]*Server
	// commandTable 指令表
	commandTable map[string]map[string]abstract.ICommand
	// 服务启停的信号
	sigChan = make(chan int, 0)
)

// NewWebsocketServe 启动websocket服务
//
// Author : go_developer@163.com<张德满>
//
// Date : 6:49 下午 2021/3/27
func NewWebsocketServe(wsInstanceList ...abstract.IWebsocket) error {
	if len(wsInstanceList) == 0 {
		return errors.WithStack(errors.New("register websocket server list is empty"))
	}
	ginRouterTable = make(map[int]*gin.Engine)
	wsServerTable = make(map[int]map[string]*Server)
	commandTable = make(map[string]map[string]abstract.ICommand)
	for _, wsInstance := range wsInstanceList {
		initServer(wsInstance)
	}
	run()
	return nil
}

// 初始化server
func initServer(wsInstance abstract.IWebsocket) {
	// 初始化ws server
	// 初始化 gin 路由表
	if _, exist := ginRouterTable[wsInstance.GetServerPort()]; !exist {
		ginRouterTable[wsInstance.GetServerPort()] = gin.Default()
	}
	// 初始化WS-Server表
	if _, exist := wsServerTable[wsInstance.GetServerPort()]; !exist {
		wsServerTable[wsInstance.GetServerPort()] = make(map[string]*Server)
	}
	if _, exist := wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()]; !exist {
		// 生成并存储 WS-Server
		wsSetConfigList := wsInstance.GetWSServerConfig()
		if nil == wsSetConfigList {
			wsSetConfigList = make([]config.SetWSServerConfig, 0)
		}
		s := &Server{
			ginRouter: gin.Default(),
			wsServer:  melody.New(),
			conf:      config.NewWSServerConfig(wsSetConfigList...),
		}
		if s.conf.LogEnable {
			// 开启了日志,初始化日志
			if len(s.conf.LogPath) == 0 {
				panic(wsInstance.GetModuleFlag() + " 模块开启了日志记录,但是没有配置日志路径")
			}
			var (
				err     error
				logConf *logger.RotateLogConfig
			)
			optionList := make([]logger.SetLoggerOptionFunc, 0)
			if s.conf.LogConsole {
				optionList = append(optionList, logger.WithConsoleOutput())
			}
			if s.conf.LogConsole {
				optionList = append(optionList, logger.WithConsoleOutput())
			}
			if len(s.conf.LogFile) == 0 {
				s.conf.LogFile = wsInstance.GetModuleFlag() + ".log"
			}
			if logConf, err = logger.NewRotateLogConfig(s.conf.LogPath, s.conf.LogFile, logger.WithTimeIntervalType(s.conf.LogSplitInterval)); nil != err {
				panic(wsInstance.GetModuleFlag() + " 模块开启了日志记录,日志初始化失败, 失败原因 : " + err.Error())
			}
			if s.loggerInstance, err = logger.NewLogger(s.conf.LogLevel, logConf, optionList...); nil != err {
				panic(wsInstance.GetModuleFlag() + " 模块开启了日志记录,日志初始化失败, 失败原因 : " + err.Error())
			}
		}
		if nil == s.loggerInstance && s.conf.LogConsole {
			var err error
			// 没有配置文件日志, 但是配置了控制台输出
			if s.loggerInstance, err = logger.NewConsoleLogger(s.conf.LogLevel); nil != err {
				panic(wsInstance.GetModuleFlag() + " 模块开启了控制台日志记录,日志初始化失败, 失败原因 : " + err.Error())
			}
		}
		wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()] = s
	}

	// 初始化指令存储表
	if _, exist := commandTable[wsInstance.GetModuleFlag()]; !exist {
		commandTable[wsInstance.GetModuleFlag()] = make(map[string]abstract.ICommand)
	}
	routerGroup := ginRouterTable[wsInstance.GetServerPort()].Group(wsInstance.GetModuleFlag())
	// 注册路由
	for _, path := range wsInstance.HandshakeURL() {
		routerGroup.GET(path, func(ctx *gin.Context) {
			parameter := map[string]interface{}{
				"ws_context": context.NewContext(ctx, wsInstance.GetModuleFlag(), nil),
			}
			_ = wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].wsServer.HandleRequestWithKeys(ctx.Writer, ctx.Request, parameter)
		})
	}
	// 注册回调函数
	// 1. 建立连接的函数
	wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].wsServer.HandleConnect(func(session *melody.Session) {
		ctxInterface, _ := session.Get("ws_context")
		ctx := ctxInterface.(*context.WSContext)
		ctx.Session = session
		if err := wsInstance.Connect(ctx); nil == err {
			if wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.StoreConnection && nil != wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.ConnectionManager {
				wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.ConnectionManager.Store(ctx)
			}
		}
	})
	// 2. 指令处理的函数
	wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].wsServer.HandleMessage(func(session *melody.Session, bytes []byte) {
		// TODO : 增加指令回调失败的callback
		ctxInterface, _ := session.Get("ws_context")
		ctx := ctxInterface.(*context.WSContext)
		_ = dispatchCommand(ctx, bytes)
	})
	// 3, 关闭连接的处理函数
	wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].wsServer.HandleClose(func(session *melody.Session, i int, s string) error {
		ctxInterface, _ := session.Get("ws_context")
		ctx := ctxInterface.(*context.WSContext)
		defer func() {
			if !wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.StoreConnection || nil == wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.ConnectionManager {
				return
			}
			wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.ConnectionManager.Del(ctx, "")
		}()
		return wsInstance.Close(ctx, i, s)
	})
	// 4. 断开连接的处理函数
	wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].wsServer.HandleDisconnect(func(session *melody.Session) {
		ctxInterface, _ := session.Get("ws_context")
		ctx := ctxInterface.(*context.WSContext)
		defer func() {
			if nil == wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.ConnectionManager {
				return
			}
			wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].conf.ConnectionManager.Del(ctx, "")
		}()
		wsInstance.Disconnect(ctx)
	})
	// 注册指令
	for _, cmd := range wsInstance.GetCommandList() {
		if nil != wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].loggerInstance {
			wsServerTable[wsInstance.GetServerPort()][wsInstance.GetModuleFlag()].loggerInstance.Debug(
				"长连接指令注册成功",
				zap.String("module", wsInstance.GetModuleFlag()),
				zap.String("command", cmd.GetCommand()),
			)
		}
		commandTable[wsInstance.GetModuleFlag()][cmd.GetCommand()] = cmd
	}
	go func() {
		if err := ginRouterTable[wsInstance.GetServerPort()].Run(fmt.Sprintf(":%d", wsInstance.GetServerPort())); nil != err {
			panic(err)
		}
	}()
}

// dispatchCommand 调度command ...
//
// Author : go_developer@163.com<张德满>
//
// Date : 3:36 下午 2021/3/28
func dispatchCommand(ctx *context.WSContext, data []byte) error {
	if _, exist := commandTable[ctx.Flag]; !exist {
		return errors.WithStack(errors.New("未注册【" + ctx.Flag + "】长连接模块"))
	}

	cmd := gjson.Get(string(data), "command").String()
	var (
		exist       bool
		cmdInstance abstract.ICommand
		cmdConfig   *config.CommandConfig
		err         error
	)
	if cmdInstance, exist = commandTable[ctx.Flag][cmd]; !exist {
		return errors.WithStack(errors.New("【" + ctx.Flag + "】长连接模块未注册【" + cmd + "】指令"))
	}
	optionList := cmdInstance.GetConfigOption()
	if nil == optionList {
		optionList = make([]config.SetCommandConfig, 0)
	}
	cmdConfig = config.NewCommandConfig(optionList...)
	if err = cmdInstance.Execute(ctx, data); nil != err {
		if cmdConfig.PushMessageWithError {
			_ = message.Response(ctx, map[string]interface{}{
				"command": cmd,
				"message": err.Error(),
				"success": false,
			})
		}
		return err
	}
	return nil
}

func run() {
	<-sigChan
	// TODO : 增加后置hook
}

// Stop 停止服务
//
// Author : go_developer@163.com<张德满>
//
// Date : 3:55 下午 2021/3/28
func Stop() {
	sigChan <- 1
}
